@@ -9,6 +9,7 @@ from datetime import datetime, timedelta
 import operator
 import sys
 import time
+import types
 import re
 
 import redis
@@ -28,7 +29,7 @@ if not PY2:
     long = int
 
 
-__version__ = '0.7.0'
+__version__ = '0.8.2'
 
 
 if sys.version_info[0] == 2:
@@ -39,8 +40,6 @@ if sys.version_info[0] == 2:
     int_to_byte = chr
 
     def to_bytes(x, charset=sys.getdefaultencoding(), errors='strict'):
-        if x is None:
-            return None
         if isinstance(x, (bytes, bytearray, buffer)) or hasattr(x, '__str__'):
             return bytes(x)
         if isinstance(x, unicode):
@@ -71,8 +70,6 @@ else:
     int_to_byte = operator.methodcaller('to_bytes', 1, 'big')
 
     def to_bytes(x, charset=sys.getdefaultencoding(), errors='strict'):
-        if x is None:
-            return None
         if isinstance(x, (bytes, bytearray, memoryview)):
             return bytes(x)
         if isinstance(x, str):
@@ -171,6 +168,39 @@ class _Hash(_StrKeyDict):
     redis_type = b'hash'
 
 
+def DecodeGenerator(gen):
+    for item in gen:
+        yield _decode(item)
+
+
+def _decode(value):
+    if isinstance(value, bytes):
+        value = value.decode()
+    elif isinstance(value, dict):
+        value = dict((_decode(k), _decode(v)) for k, v in value.items())
+    elif isinstance(value, (list, set, tuple)):
+        value = value.__class__(_decode(x) for x in value)
+    elif isinstance(value, types.GeneratorType):
+        value = DecodeGenerator(value)
+    return value
+
+
+def _make_decode_func(func):
+    def decode_response(*args, **kwargs):
+        val = _decode(func(*args, **kwargs))
+        return val
+    return decode_response
+
+
+def _patch_responses(obj):
+    for attr_name in dir(obj):
+        attr = getattr(obj, attr_name)
+        if not callable(attr) or attr_name.startswith('_'):
+            continue
+        func = _make_decode_func(attr)
+        setattr(obj, attr_name, func)
+
+
 class FakeStrictRedis(object):
     @classmethod
     def from_url(cls, url, db=None, **kwargs):
@@ -180,9 +210,10 @@ class FakeStrictRedis(object):
                 db = int(url.path.replace('/', ''))
             except (AttributeError, ValueError):
                 db = 0
-        return cls(db=db)
+        return cls(db=db, **kwargs)
 
-    def __init__(self, db=0, charset='utf-8', errors='strict', **kwargs):
+    def __init__(self, db=0, charset='utf-8', errors='strict',
+                 decode_responses=False, **kwargs):
         if db not in DATABASES:
             DATABASES[db] = _StrKeyDict()
         self._db = DATABASES[db]
@@ -190,6 +221,9 @@ class FakeStrictRedis(object):
         self._encoding = charset
         self._encoding_errors = errors
         self._pubsubs = []
+        self._decode_responses = decode_responses
+        if decode_responses:
+            _patch_responses(self)
 
     def flushdb(self):
         DATABASES[self._db_num].clear()
@@ -231,17 +265,33 @@ class FakeStrictRedis(object):
     __contains__ = exists
 
     def expire(self, name, time):
+        return self._expire(name, time)
+
+    def pexpire(self, name, millis):
+        return self._expire(name, millis, 1000)
+
+    def _expire(self, name, time, multiplier=1):
         if isinstance(time, timedelta):
-            time = int(timedelta_total_seconds(time))
+            time = int(timedelta_total_seconds(time) * multiplier)
+        if not isinstance(time, int):
+            raise redis.ResponseError("value is not an integer or out of "
+                                      "range.")
         if self.exists(name):
-            self._db.expire(name, datetime.now() + timedelta(seconds=time))
+            self._db.expire(name, datetime.now() +
+                            timedelta(seconds=time / float(multiplier)))
             return True
         else:
             return False
 
     def expireat(self, name, when):
+        return self._expireat(name, when)
+
+    def pexpireat(self, name, when):
+        return self._expireat(name, when, 1000)
+
+    def _expireat(self, name, when, multiplier=1):
         if not isinstance(when, datetime):
-            when = datetime.fromtimestamp(when)
+            when = datetime.fromtimestamp(when / float(multiplier))
         if self.exists(name):
             self._db.expire(name, when)
             return True
@@ -262,7 +312,7 @@ class FakeStrictRedis(object):
             return to_bytes(value)
 
     def __getitem__(self, name):
-        return self._db[name]
+        return self.get(name)
 
     def getbit(self, name, offset):
         """Returns a boolean indicating the value of ``offset`` in ``name``"""
@@ -379,11 +429,23 @@ class FakeStrictRedis(object):
     def set(self, name, value, ex=None, px=None, nx=False, xx=False):
         if (not nx and not xx) or (nx and self._db.get(name, None) is None) \
                 or (xx and not self._db.get(name, None) is None):
-            if ex is not None and ex > 0:
-                self._db.expire(name, datetime.now() + timedelta(seconds=ex))
-            elif px is not None and px > 0:
-                self._db.expire(name, datetime.now() +
-                                timedelta(milliseconds=px))
+            if ex is not None:
+                if isinstance(ex, timedelta):
+                    ex = ex.seconds + ex.days * 24 * 3600
+                if ex < 0:
+                    raise ResponseError('invalid expire time in SETEX')
+                if ex > 0:
+                    self._db.expire(name, datetime.now() +
+                                    timedelta(seconds=ex))
+            elif px is not None:
+                if isinstance(px, timedelta):
+                    ms = int(px.microseconds / 1000)
+                    px = (px.seconds + px.days * 24 * 3600) * 1000 + ms
+                if px < 0:
+                    raise ResponseError('invalid expire time in SETEX')
+                if px > 0:
+                    self._db.expire(name, datetime.now() +
+                                    timedelta(milliseconds=px))
             self._db[name] = to_bytes(value)
             return True
         else:
@@ -458,19 +520,19 @@ class FakeStrictRedis(object):
 
     def _ttl(self, name, multiplier=1):
         if name not in self._db:
-            return None
+            return -2
 
         exp_time = self._db.expiring(name)
         if not exp_time:
-            return None
+            return -1
 
         now = datetime.now()
         if now > exp_time:
             return None
         else:
-            return round(((exp_time - now).days * 3600 * 24
-                          + (exp_time - now).seconds
-                          + (exp_time - now).microseconds / 1E6) * multiplier)
+            return long(round(((exp_time - now).days * 3600 * 24 +
+                        (exp_time - now).seconds +
+                        (exp_time - now).microseconds / 1E6) * multiplier))
 
     def type(self, name):
         key = self._db.get(name)
@@ -694,6 +756,7 @@ class FakeStrictRedis(object):
     def rpoplpush(self, src, dst):
         el = self.rpop(src)
         if el is not None:
+            el = to_bytes(el)
             try:
                 self._db[dst].insert(0, el)
             except KeyError:
@@ -727,6 +790,7 @@ class FakeStrictRedis(object):
     def brpoplpush(self, src, dst, timeout=0):
         el = self.rpop(src)
         if el is not None:
+            el = to_bytes(el)
             try:
                 self._db[dst].insert(0, el)
             except KeyError:
@@ -828,7 +892,7 @@ class FakeStrictRedis(object):
 
     def hvals(self, name):
         "Return the list of values within hash ``name``"
-        return self._db.get(name, {}).values()
+        return list(self._db.get(name, {}).values())
 
     def sadd(self, name, *values):
         "Add ``value`` to set ``name``"
@@ -855,7 +919,7 @@ class FakeStrictRedis(object):
         set named ``dest``.  Returns the number of keys in the new set.
         """
         diff = self.sdiff(keys, *args)
-        self._db[dest] = diff
+        self._db[dest] = set(to_bytes(x) for x in diff)
         return len(diff)
 
     def sinter(self, keys, *args):
@@ -872,7 +936,7 @@ class FakeStrictRedis(object):
         set named ``dest``.  Returns the number of keys in the new set.
         """
         intersect = self.sinter(keys, *args)
-        self._db[dest] = intersect
+        self._db[dest] = set(to_bytes(x) for x in intersect)
         return len(intersect)
 
     def sismember(self, name, value):
@@ -927,7 +991,7 @@ class FakeStrictRedis(object):
         set named ``dest``.  Returns the number of keys in the new set.
         """
         union = self.sunion(keys, *args)
-        self._db[dest] = union
+        self._db[dest] = set(to_bytes(x) for x in union)
         return len(union)
 
     def _get_zelement_range_filter_func(self, min_val, max_val):
@@ -1419,11 +1483,12 @@ class FakeStrictRedis(object):
                     continue
         raise redis.WatchError('Could not run transaction after 5 tries')
 
-    def pubsub(self):
+    def pubsub(self, ignore_subscribe_messages=False):
         """
         Returns a new FakePubSub instance
         """
-        ps = FakePubSub()
+        ps = FakePubSub(decode_responses=self._decode_responses,
+                        ignore_subscribe_messages=ignore_subscribe_messages)
         self._pubsubs.append(ps)
 
         return ps
@@ -1553,6 +1618,14 @@ class FakeRedis(FakeStrictRedis):
             raise redis.RedisError("ZADD is missing kwargs param")
         return super(FakeRedis, self).zadd(name, **pairs)
 
+    def ttl(self, name):
+        r = super(FakeRedis, self).ttl(name)
+        return r if r >= 0 else None
+
+    def pttl(self, name):
+        r = super(FakeRedis, self).pttl(name)
+        return r if r >= 0 else None
+
 
 class FakePipeline(object):
     """Helper class for FakeStrictRedis to implement pipelines.
@@ -1655,14 +1728,15 @@ class FakePubSub(object):
     UNSUBSCRIBE_MESSAGE_TYPES = ['unsubscribe', 'punsubscribe']
     PATTERN_MESSAGE_TYPES = ['psubscribe', 'punsubscribe']
 
-    def __init__(self, *args, **kwargs):
+    def __init__(self, decode_responses=False, *args, **kwargs):
         self.channels = {}
         self.patterns = {}
         self._q = Queue()
         self.subscribed = False
-
-        self.ignore_subscribe_messages = kwargs['ignore_subscribe_messages']\
-            if 'ignore_subscribe_messages' in kwargs else False
+        if decode_responses:
+            _patch_responses(self)
+        self.ignore_subscribe_messages = kwargs.get(
+            'ignore_subscribe_messages', False)
 
     def put(self, channel, message, message_type, pattern=None):
         """
